import os.path
import datetime
import base64
from paradict import errors, misc, const, kv, xtypes
from paradict.const import Datatype
from paradict.typeref import TypeRef
from paradict.deserializer.buffered_text_stream import BufferedTextStream

__all__ = ["Decoder"]


class Decoder:
    """Class to convert some textual Paradict data into a Python dict"""
    def __init__(self, *, type_ref=None, receiver=None, obj_builder=None,
                 root_dir=None):
        """
        Init

        [param]
        - type_ref: optional TypeRef object
        - receiver: callback function that will be called at the end of conversion.
        This callback function accepts the Decoder instance as argument
        - obj_builder: function that accepts a paradict.xtypes.Obj container and
        returns a fresh new Python object
        - root_dir: root directory in which the attachments dir is supposed to be
        """
        self._type_ref = type_ref if type_ref else TypeRef()
        self._receiver = receiver
        self._obj_builder = obj_builder
        self._root_dir = root_dir
        # data
        self._data = dict()
        # misc
        self._queue = BufferedTextStream()
        self._stack = list()
        self._lineno = 1
        self._active = False
        self._ordered_converters = (self._decode_null,
                                    self._decode_bool, self._decode_str,
                                    self._decode_complex_number,
                                    self._decode_int, self._decode_float,
                                    self._decode_date, self._decode_time,
                                    self._decode_datetime, self._decode_load_func)

    @property
    def data(self):
        if self._stack:
            context = self._stack[0]
            return context.container
        return self._data

    @property
    def type_ref(self):
        return self._type_ref

    @type_ref.setter
    def type_ref(self, val):
        self._type_ref = val

    @property
    def receiver(self):
        return self._receiver

    @receiver.setter
    def receiver(self, val):
        self._receiver = val

    @property
    def obj_builder(self):
        return self._obj_builder

    @obj_builder.setter
    def obj_builder(self, val):
        self._obj_builder = val

    @property
    def root_dir(self):
        return self._root_dir

    @root_dir.setter
    def root_dir(self, val):
        self._root_dir = val

    @property
    def queue(self):
        return self._queue

    @queue.setter
    def queue(self, val):
        self._queue = val

    def feed(self, s):
        """
        Feed the decoder engine with some string.
        The string might represent a line in the textual Paradict data,
        or an arbitrary length of characters.

        Note: it is very important to make sure that each line is ended
        by a "\n" newline character

        Check the `paradict.decode` function to see an example of
        how to use this class
        """
        self._queue.put(s)
        for line in self._queue.get_all():
            line = line.rstrip("\n")
            if line.rstrip() == "===":  # END
                self._cleanup_stack(0)
                self._stack = list()
                self._active = False
                if self._receiver:
                    self._receiver(self)
                return
            if not self._active:
                self._data = self._type_ref.dict_type()
                #root_context = Context("dict", self._data, 0)
                #self._stack.append(root_context)
                self._update_stack("dict", self._data, 0)
                self._active = True
            self._process(line)
        return

    def _process(self, line):
        line = self._check_line(line)
        # process line
        try:
            self._interpret(line)
        except Exception as e:
            msg = "Error on line {}: {}{}\n--> {}"
            str_e = str(e)
            str_e = ": " + str_e if str_e else str_e
            msg = msg.format(self._lineno, type(e).__name__, str_e, line)
            raise type(e)(msg)
        self._lineno += 1

    def _interpret(self, line):
        #if not self._stack:
        #    return
        context = self._get_context()
        name = context.name
        container = context.container
        if name in ("dict", "list", "set", "obj"):
            self._update_context(line)
        elif name in ("bin", "raw", "text", "float", "int", "grid"):
            container.append(line)
        else:
            msg = "Unknown tag '{}'. Expected dict, list, set, obj, grid, bin, str, float, int, or raw."
            msg = msg.format(name)
            raise errors.Error(msg)

    def _update_context(self, line):
        context = self._get_context()
        name = context.name
        updaters = {"dict": self._update_dict_container,
                    "list": self._update_list_container,
                    "set": self._update_set_container,
                    "obj": self._update_obj_container}
        updater = updaters.get(name)
        if not updater:
            raise errors.Error
        updater(context, line)

    def _update_dict_container(self, context, line):
        container = context.container
        indents = context.indents
        # process whitespaces
        if not line or line.isspace():
            pass
        # ignore comments
        elif line.startswith("#"):
            return
        # process key value
        else:
            info = kv.split(line)
            key, val, mode = info.key, info.val, info.mode
            key = self._check_key(key, mode)
            value = self._decode_value(val)
            tag = self._check_multiline_tag(val)
            if tag in ("dict", "list", "set"):
                container[key] = value
                context.stored_dict_key = None
                self._update_stack(tag, value, indents + 1)
            elif tag in ("grid", "obj", "bin", "raw", "text", "int", "float"):
                context.stored_dict_key = key
                self._update_stack(tag, value, indents + 1)
            else:
                container[key] = value

    def _update_obj_container(self, context, line):
        return self._update_dict_container(context, line)

    def _update_list_container(self, context, line):
        container = context.container
        indents = context.indents
        # process whitespaces
        if not line or line.isspace():
            pass
        # ignore comments
        elif line.startswith("#"):
            return
        # process key value
        else:
            value = self._decode_value(line)
            tag = self._check_multiline_tag(line)
            if tag in ("dict", "list", "set"):
                container.append(value)
                self._update_stack(tag, value, indents + 1)
            elif tag in ("grid", "obj", "bin", "raw", "text", "int", "float"):
                self._update_stack(tag, value, indents + 1)
            else:
                container.append(value)

    def _update_set_container(self, context, line):
        container = context.container
        indents = context.indents
        # process whitespaces
        if not line or line.isspace():
            pass
        # ignore comments
        elif line.startswith("#"):
            return
        # process key value
        else:
            value = self._decode_value(line)
            tag = self._check_multiline_tag(line)
            if tag in ("bin", "raw", "text", "float", "int"):
                self._update_stack(tag, value, indents + 1)
            elif tag in ("dict", "list", "set", "grid", "obj"):
                msg = "Set can't contain any of: dict, list, set, grid, obj"
                raise errors.Error(msg)
            else:
                container.add(value)

    def _check_line(self, line):
        """Check the validity of the indent in the line
        Returns the indent-less version of line"""
        if not line:
            return line
        context = self._get_context()
        strict = False if context.name in ("bin", "raw", "text", "int", "float") else True
        indents = misc.count_indents(line, strict)
        # same indent
        if indents == context.indents:
            pass
        # lower indent
        elif indents < context.indents:
            self._cleanup_stack(indents)
        # exaggerated forward indent
        elif indents > context.indents and strict:
            if context.indents == 0:
                msg = "Expected {} indent at line {}:{}"
            else:
                msg = "Expected {} or less indents at line {}:{}"
            msg = msg.format(context.indents, self._lineno, line)
            raise errors.IndentError(msg)
        return misc.dedent(line, context.indents)

    def _check_key(self, key, mode):
        if mode == const.CONFIG_MODE:
            if " " in key or "=" in key:
                msg = ("Keys in CONFIG_MODE are strings that "
                       "shouldn't contain a space "
                       "or the equal sign (=).")
                raise errors.Error(msg)
            key = "'{}'".format(key)
        key = self._decode_value(key)
        if self._type_ref.check(type(key)) in (Datatype.STR, Datatype.INT,
                                               Datatype.FLOAT,
                                               Datatype.COMPLEX):
            return key
        # exception
        if mode == const.DATA_MODE:
            msg = ("In DATA_MODE, a key should be one of: str, "
                   "int, float, complex.")
        else:
            msg = ("Keys in CONFIG_MODE are strings that "
                   "shouldn't contain a space "
                   "or the equal sign (=).")
        raise errors.Error(msg)

    def _check_multiline_tag(self, value):
        if not value.startswith("(") or not value.endswith(")"):
            return None
        value = value.strip("()")
        if value not in ("bin", "dict", "list", "set",
                         "raw", "text", "int", "float", "obj", "grid"):
            return None
        return value

    def _update_stack(self, container_name,
                      container, indents):
        new_context = Context(container_name, container, indents)
        self._stack.append(new_context)

    def _cleanup_stack(self, indents):
        if not self._stack:
            return
        while True:
            if len(self._stack) == 1:
                break
            context = self._get_context()
            parent_context = self._get_parent_context()
            if indents >= context.indents:
                break
            # bin
            if context.name == "bin":
                self._consume_bin_block(parent_context, context)
            # raw
            elif context.name == "raw":
                self._consume_raw_block(parent_context, context)
            # multiline str
            elif context.name == "text":
                self._consume_str_block(parent_context, context)
            # obj
            elif context.name == "obj":
                self._consume_obj_block(parent_context, context)
            # obj
            elif context.name == "grid":
                self._consume_grid_block(parent_context, context)
            # delete the latest context
            del self._stack[-1]

    def _consume_bin_block(self, parent_context, context):
        bin_block = context.container
        val = list()
        for line in bin_block:
            if not line or line.isspace():
                continue
            for char in line:
                if not char.isspace():
                    val.append(char)
        #value = [item.strip() for item in bin_block if item]
        value = "".join(val)
        bin_data = self._decode_bin(value)
        #default_bin_type = self._bin_type
        #if type(bin_data) is not default_bin_type:
        #    bin_data = default_bin_type(bin_data)
        self._update_parent_context(parent_context, bin_data)

    def _consume_raw_block(self, parent_context, context):
        # concatenate str lines
        block = misc.strip_block_extra_space(context.container)
        text = "'{}'".format(block)
        text = self._decode_str(text)
        self._update_parent_context(parent_context, text)

    def _consume_str_block(self, parent_context, context):
        # concatenate text lines
        block = misc.strip_block_extra_space(context.container)
        text = '"{}"'.format(block)
        text = self._decode_str(text)
        self._update_parent_context(parent_context, text)

    def _consume_obj_block(self, parent_context, context):
        container = context.container
        if self._obj_builder:
            container = self._obj_builder(container)
        self._update_parent_context(parent_context, container)

    def _consume_grid_block(self, parent_context, context):
        container = context.container
        row_size = 0
        grid = list()
        # turn lines into rows
        for line in container:
            if not line or line.isspace():
                continue
            temp_row = line.split()
            if row_size == 0:
                row_size = len(temp_row)
            elif len(temp_row) != row_size:
                msg = "Inconsistent grid"
                raise errors.Error(msg)
            row = list()
            for item in temp_row:
                val = self._decode_value(item)
                dtype = self._type_ref.check(type(val))
                if dtype not in (Datatype.INT, Datatype.FLOAT, Datatype.COMPLEX):
                    msg = "A grid should be built with int, float, or complex numbers"
                    raise errors.Error(msg)
                row.append(val)
            grid.append(tuple(row))
        # update parent context
        grid = self._type_ref.grid_type(grid)
        self._update_parent_context(parent_context, grid)

    def _update_parent_context(self, parent_context, data):
        if parent_context.name in ("dict", "obj"):
            key = parent_context.stored_dict_key
            parent_context.container[key] = data
        elif parent_context.name == "list":
            parent_context.container.append(data)
        elif parent_context.name == "set":
            parent_context.container.add(data)
        else:
            msg = "Invalid parent name '{}'"
            msg = msg.format(parent_context.name)
            raise errors.Error(msg)

    def _get_context(self):
        try:
            return self._stack[-1]
        except IndexError as e:
            msg = "Inconsistent stack state"
            raise errors.Error(msg)

    def _get_parent_context(self):
        try:
            return self._stack[-2]
        except IndexError as e:
            msg = "Inconsistent stack state"
            raise errors.Error(msg)

    def _decode_value(self, value):
        value = value.strip()
        container = self._decode_container(value)
        if container is not None:
            return container
        # scalars and co.  (note that the order of items matters here)
        for p in self._ordered_converters:
            try:
                result = p(value)
            except Exception as e:
                continue
            else:
                return result
        raise errors.Error("Conversion error")

    def _decode_container(self, val):
        if not val.startswith("(") or not val.endswith(")"):
            return None
        # bin
        if val == "(bin)":
            return list()
        # dict
        if val == "(dict)":
            return self._type_ref.dict_type()
        # list
        if val == "(list)":
            return self._type_ref.list_type()
        # set
        if val == "(set)":
            return self._type_ref.set_type()
        # obj
        if val == "(obj)":
            return self._type_ref.obj_type()
        # raw
        if val == "(raw)":
            return list()
        # multiline string
        if val == "(text)":
            return list()
        # multiline float
        if val == "(float)":
            return list()
        # multiline int
        if val == "(int)":
            return list()
        # grid
        if val == "(grid)":
            return list()
        msg = ("Expected one of these tags: (bin), "
               "(dict), (list), (set), (raw), (text), (int), (float), (obj), (grid)")
        raise errors.Error(msg)

    def _decode_null(self, val):
        return decode_null(val)

    def _decode_bool(self, val):
        r = decode_bool(val)
        return self._type_ref.bool_type(r)

    def _decode_bin(self, val):
        r = base64.b16decode(val, casefold=True)
        r = r if self._type_ref.bin_type is bytes else self._type_ref.bin_type(r)
        return r

    def _decode_str(self, val):
        r = decode_str(val)
        r = r if self._type_ref.str_type is str else self._type_ref.str_type(r)
        return r

    def _decode_complex_number(self, val):
        r = decode_complex(val)
        r = r if self._type_ref.complex_type is complex else self._type_ref.complex_type(r)
        return r

    def _decode_int(self, val):
        r = decode_int(val)
        r = r if self._type_ref.int_type is int else self._type_ref.int_type(r)
        return r

    def _decode_float(self, val):
        r = decode_float(val)
        r = r if self._type_ref.float_type is float else self._type_ref.float_type(r)
        return r

    def _decode_date(self, val):
        r = decode_date(val)
        r = r if self._type_ref.date_type is datetime.date else self._type_ref.date_type(r)
        return r

    def _decode_time(self, val):
        r = decode_time(val)
        r = r if self._type_ref.time_type is datetime.time else self._type_ref.time_type(r)
        return r

    def _decode_datetime(self, val):
        r = decode_datetime(val)
        r = r if self._type_ref.datetime_type is datetime.datetime else self._type_ref.datetime_type(r)
        return r

    def _decode_load_func(self, val):
        filename = decode_load_func(val)
        filename = filename.strip("'\"`")
        if not os.path.isabs(filename):
            parts = misc.split_relative_path(filename)
            root_dir = self._root_dir if self._root_dir else os.getcwd()
            filename = os.path.join(root_dir, *parts)
        if not os.path.isfile(filename):
            msg = "Attachment not found. Filename: {}".format(filename)
            raise errors.Error(msg)
        with open(filename, "rb") as file:
            r = file.read()
        r = r if self._type_ref.bin_type is bytes else self._type_ref.bin_type(r)
        return r


class Context:
    def __init__(self, name, container, indents):
        self._name = name
        self._container = container
        self._indents = indents
        self._stored_dict_key = None
        self._is_dict_key_stored = False

    @property
    def name(self):
        return self._name

    @property
    def container(self):
        return self._container

    @property
    def indents(self):
        return self._indents

    @property
    def stored_dict_key(self):
        return self._stored_dict_key

    @stored_dict_key.setter
    def stored_dict_key(self, val):
        self._stored_dict_key = val

    @property
    def is_dict_key_stored(self):
        return self._is_dict_key_stored

    @is_dict_key_stored.setter
    def is_dict_key_stored(self, val):
        self._is_dict_key_stored = val


# === Decoders ===


def decode_bin(val):
    # list of strings
    val = "".join(val)
    return base64.standard_b64decode(val)


def decode_bool(val):
    val = val.lower()
    if val == "true":
        return bool(1)
    elif val == "false":
        return bool(0)
    raise errors.Error


def decode_complex(val):
    if val.endswith("i"):
        val = val.replace(" ", "")
        val = val.rstrip("i") + "j"
        return complex(val)
    raise errors.Error


def decode_date(val):
    """Parse date and time with the ISO 8601 format"""
    # date ISO format
    return datetime.date.fromisoformat(val)


def decode_datetime(val):
    # datetime (2023-03-13T09:10:42Z)
    if val.endswith("Z"):
        val = val.rstrip("Z")
        val += "+00:00"
    return datetime.datetime.fromisoformat(val)


def decode_float(val):
    val = val.lower()
    if "e" in val or "." in val:
        return float(val)
    raise errors.Error


def decode_int(val):
    v = val.lstrip("+-")
    # hexadecimal
    if v.startswith("0x"):
        val = int(val, 16)
        return xtypes.HexInt(val)
    # octal
    if v.startswith("0o"):
        val = int(val, 8)
        return xtypes.OctInt(val)
    # binary
    if v.startswith("0b"):
        val = int(val, 2)
        return xtypes.BinInt(val)
    # default base 10
    return int(val)


def decode_null(val):
    val = val.lower()
    if val == "null":
        return None
    raise errors.Error


def decode_str(val):
    if val.startswith('"') and val.endswith('"'):
        return misc.decode_unicode(val[1:-1])
    if val.startswith("'") and val.endswith("'"):
        return val[1:-1]
    raise errors.Error("Conversion error")


def decode_time(val):
    if val.endswith("Z"):
        val = val.rstrip("Z")
        val += "+00:00"
    # time ISO format
    return datetime.time.fromisoformat(val)


def decode_load_func(val):
    a, b = "load(", ")"
    if val.startswith(a) and val.endswith(b):
        argument = val[len(a):-1]
        return decode_str(argument)
    raise errors.Error("Conversion error")



# general constants
from enum import unique, Enum


@unique
class Datatype(Enum):
    DICT = 1
    LIST = 2
    SET = 3
    OBJ = 4
    GRID = 5
    BOOL = 6
    STR = 7
    BIN = 8
    INT = 9
    FLOAT = 10
    COMPLEX = 11
    DATE = 12
    TIME = 13
    DATETIME = 14


# Indent
INDENT_WIDTH = 4  # Four spaces
INDENT_STR = " " * INDENT_WIDTH

# Paradict Epoch: 2020-01-01T00:00:00Z
COVID_YEAR = 2020

# modes for Textual Paradict
DATA_MODE = "d"
CONFIG_MODE = "c"



# File generated by Setupinit
import sys
import paradict

__all__ = []

import paradict.io_bin

import paradict.io_text

TEXT = """\
Paradict - Streamable multi-format serialization
https://github.com/pyrustic/paradict

COMMANDS:
    t2b <src> [<dst>]: Text-to-binary serialization.
    b2t <src> [<dst>]: Binary-to-text serialization.

NOTES:
    - Serialization is done from a src file to a dst file.
    - <src> is an existing filename.
    - <dst> is an optional filename (that exists or not).

To show this help message, type '-h' or '--help'.
"""


def main():
    all_args = sys.argv[1:]
    try:
        command = all_args[0]
    except IndexError as e:
        print(TEXT)
        return
    args = all_args[1:]
    if command in ("-h", "--help"):
        print(TEXT)
        return
    handler = HANDLERS.get(command)
    if handler is None:
        print("Invalid command. Type -h for help.")
    else:
        handler(args)


def t2b(args):
    args = ensure_args(args)
    if not args:
        return
    src, dst = args
    data = paradict.io_text.read(src)
    if dst:
        paradict.io_bin.pack_into(data, dst)
    else:
        print(paradict.pack(data))


def b2t(args):
    args = ensure_args(args)
    if not args:
        return
    src, dst = args
    data = paradict.io_bin.unpack_from(src)
    if dst:
        paradict.io_text.write(data, dst)
    else:
        print(paradict.encode(data))


def ensure_args(args):
    if not args:
        print("Error: The 'src' argument is required but was not provided.")
    elif len(args) == 1:
        return args[0], None
    elif len(args) == 2:
        return args
    else:
        print("Error: Too many arguments.")


HANDLERS = {"t2b": t2b, "b2t": b2t}


if __name__ == "__main__":
    main()



"""Boxes to hold grids, hexadecimal integers, et cetera"""
from collections import UserList, UserDict
from paradict import misc


__all__ = ["Grid", "Obj", "HexInt", "OctInt", "BinInt"]


class Grid(UserList):
    """
    Box to hold a grid. A grid is made of numbers and should
    be consistent (rows should be of same size)

    Example:
    ```
    # a grid with 2 rows and 3 columns
    my_grid = Grid([(0, 1, 0),
                    (1, 0, 1)])
    ```
    """
    pass


class Obj(UserDict):
    """
    Box to hold an Extension Object. Such objects behave like a dictionary.
    An object builder is passed as argument to the right functions, thus,
    the object is consumed/used to build a new valid data value
    """
    pass


class HexInt(int):
    """Box to hold hexadecimal integer"""

    def __new__(cls, x):
        width = 0
        if isinstance(x, str):
            parts = misc.split_int(x)
            if parts.prefix == "0x":
                width = len(parts.leading_zeros) + len(parts.val)
        instance = super().__new__(cls, str(x), base=misc.get_int_base(x))
        x = hex(instance)
        x = misc.left_pad_int(x, width) if width else x
        instance.__x = misc.tidy_up_int(x, width=4)
        return instance

    def __str__(self):
        return self.__x


class OctInt(int):
    """Box to hold octal integer"""

    def __new__(cls, x):
        width = 0
        if isinstance(x, str):
            parts = misc.split_int(x)
            if parts.prefix == "0o":
                width = len(parts.leading_zeros) + len(parts.val)
        instance = super().__new__(cls, str(x), base=misc.get_int_base(x))
        x = oct(instance)
        x = misc.left_pad_int(x, width) if width else x
        instance.__x = misc.tidy_up_int(x, width=3)
        return instance

    def __str__(self):
        return self.__x


class BinInt(int):
    """Box to hold binary integer"""

    def __new__(cls, x):
        width = 0
        if isinstance(x, str):
            parts = misc.split_int(x)
            if parts.prefix == "0b":
                width = len(parts.leading_zeros) + len(parts.val)
        instance = super().__new__(cls, str(x), base=misc.get_int_base(x))
        x = bin(instance)
        x = misc.left_pad_int(x, width) if width else x
        instance.__x = misc.tidy_up_int(x, width=4)
        return instance

    def __str__(self):
        return self.__x



"""Load and dump binary Paradict from/to file"""
from paradict.serializer.packer import Packer
from paradict.deserializer.unpacker import Unpacker


__all__ = ["unpack_from", "pack_into"]


def unpack_from(file, type_ref=None, receiver=None,
                obj_builder=None):
    """
    Open a binary Paradict file then unpack its contents into Python dict

    [param]
    - file: bin file object
    - type_ref: optional TypeRef object
    - receiver: callback function that will be called at the end of conversion.
    This callback function accepts the Decoder instance as argument
    - obj_builder: function that accepts a paradict.xtypes.Obj container and
    returns a fresh new Python object

    [return]
    Return the newly built Python object
    """
    unpacker = Unpacker(type_ref=type_ref,
                        receiver=receiver,
                        obj_builder=obj_builder)
    chunk_size = 1024
    while True:
        r = file.read(chunk_size)
        if not r:
            break
        unpacker.feed(r)
    return unpacker.data


def pack_into(data, file, *, type_ref=None):
    """
    Serialize a Python data object with the Paradict binary format
    then dump it in a file

    [param]
    - data: Python data object
    - file: binary file object
    - type_ref: optional TypeRef object
    - dict_only: boolean to enforce dict as root
    """
    packer = Packer(type_ref=type_ref)
    for r in packer.pack(data):
        file.write(r)



"""Private miscellaneous module for the Paradict binary format"""
from paradict import tags


# Mapping payload sizes in bytes to PINT tags
SIZE_TO_PINT = {1: tags.PINT_8, 2: tags.PINT_16,
                3: tags.PINT_24, 4: tags.PINT_32,
                5: tags.PINT_40, 6: tags.PINT_48,
                7: tags.PINT_56, 8: tags.PINT_64}

# Mapping PINT tags to payload sizes in bytes
PINT_TO_SIZE = {tags.PINT_8: 1, tags.PINT_16: 2,
                tags.PINT_24: 3, tags.PINT_32: 4,
                tags.PINT_40: 5, tags.PINT_48: 6,
                tags.PINT_56: 7, tags.PINT_64: 8}


# Mapping payload sizes in bytes to NINT tags
SIZE_TO_NINT = {1: tags.NINT_8, 2: tags.NINT_16,
                3: tags.NINT_24, 4: tags.NINT_32,
                5: tags.NINT_40, 6: tags.NINT_48,
                7: tags.NINT_56, 8: tags.NINT_64}

# Mapping NINT tags to payload sizes in bytes
NINT_TO_SIZE = {tags.NINT_8: 1, tags.NINT_16: 2,
                tags.NINT_24: 3, tags.NINT_32: 4,
                tags.NINT_40: 5, tags.NINT_48: 6,
                tags.NINT_56: 7, tags.NINT_64: 8}


# Mapping payload sizes in bytes to fixed-length STR tags
SIZE_TO_STR = {1: tags.STR_8, 2: tags.STR_16,
               3: tags.STR_24, 4: tags.STR_32,
               5: tags.STR_40, 6: tags.STR_48,
               7: tags.STR_56, 8: tags.STR_64,
               9: tags.STR_72, 10: tags.STR_80,
               11: tags.STR_88, 12: tags.STR_96,
               13: tags.STR_104, 14: tags.STR_112,
               15: tags.STR_120, 16: tags.STR_128,
               17: tags.STR_136, 18: tags.STR_144,
               19: tags.STR_152, 20: tags.STR_160,
               21: tags.STR_168, 22: tags.STR_176,
               23: tags.STR_184, 24: tags.STR_192,
               25: tags.STR_200, 26: tags.STR_208,
               27: tags.STR_216, 28: tags.STR_224,
               29: tags.STR_232, 30: tags.STR_240,
               31: tags.STR_248, 32: tags.STR_256}

# Mapping fixed-length STR tags to payload sizes in bytes
STR_TO_SIZE = {tags.STR_8: 1, tags.STR_16: 2,
               tags.STR_24: 3, tags.STR_32: 4,
               tags.STR_40: 5, tags.STR_48: 6,
               tags.STR_56: 7, tags.STR_64: 8,
               tags.STR_72: 9, tags.STR_80: 10,
               tags.STR_88:11, tags.STR_96: 12,
               tags.STR_104: 13, tags.STR_112: 14,
               tags.STR_120: 15, tags.STR_128: 16,
               tags.STR_136: 17, tags.STR_144: 18,
               tags.STR_152: 19, tags.STR_160: 20,
               tags.STR_168: 21, tags.STR_176: 22,
               tags.STR_184: 23, tags.STR_192: 24,
               tags.STR_200: 25, tags.STR_208: 26,
               tags.STR_216: 27, tags.STR_224: 28,
               tags.STR_232: 29, tags.STR_240: 30,
               tags.STR_248: 31, tags.STR_256: 32}


# Mapping BIN tags to number of bytes for encoding payload sizes
BIN_TO_SIZE = {tags.BIN_SHORT: 1,
               tags.BIN_MEDIUM: 2,
               tags.BIN_LONG: 3,
               tags.BIN_BIG: 4,
               tags.BIN_HEAVY: 5}

# Mapping number of bytes for encoding payload sizes to BIN tags
SIZE_TO_BIN = {1: tags.BIN_SHORT,
               2: tags.BIN_MEDIUM,
               3: tags.BIN_LONG,
               4: tags.BIN_BIG,
               5: tags.BIN_HEAVY}

# Mapping variable-length STR tags to number of bytes for encoding payload sizes
VARSTR_TO_SIZE = {tags.STR_SHORT: 1,
                  tags.STR_MEDIUM: 2,
                  tags.STR_LONG: 3,
                  tags.STR_BIG: 4,
                  tags.STR_HEAVY: 5}

# Mapping number of bytes for encoding payload sizes to variable-length STR tags 
SIZE_TO_VARSTR = {1: tags.STR_SHORT,
                  2: tags.STR_MEDIUM,
                  3: tags.STR_LONG,
                  4: tags.STR_BIG,
                  5: tags.STR_HEAVY}


# Mapping lower and upper alphabet letters to CHAR tags
LETTER_TO_TAG = {"a": tags.CHAR_A, "b": tags.CHAR_B,
                 "c": tags.CHAR_C, "d": tags.CHAR_D,
                 "e": tags.CHAR_E, "f": tags.CHAR_F,
                 "g": tags.CHAR_G, "h": tags.CHAR_H,
                 "i": tags.CHAR_I, "j": tags.CHAR_J,
                 "k": tags.CHAR_K, "l": tags.CHAR_L,
                 "m": tags.CHAR_M, "n": tags.CHAR_N,
                 "o": tags.CHAR_O, "p": tags.CHAR_P,
                 "q": tags.CHAR_Q, "r": tags.CHAR_R,
                 "s": tags.CHAR_S, "t": tags.CHAR_T,
                 "u": tags.CHAR_U, "v": tags.CHAR_V,
                 "w": tags.CHAR_W, "x": tags.CHAR_X,
                 "y": tags.CHAR_Y, "z": tags.CHAR_Z,
                 "A": tags.CHAR_UP_A, "B": tags.CHAR_UP_B,
                 "C": tags.CHAR_UP_C, "D": tags.CHAR_UP_D,
                 "E": tags.CHAR_UP_E, "F": tags.CHAR_UP_F,
                 "G": tags.CHAR_UP_G, "H": tags.CHAR_UP_H,
                 "I": tags.CHAR_UP_I, "J": tags.CHAR_UP_J,
                 "K": tags.CHAR_UP_K, "L": tags.CHAR_UP_L,
                 "M": tags.CHAR_UP_M, "N": tags.CHAR_UP_N,
                 "O": tags.CHAR_UP_O, "P": tags.CHAR_UP_P,
                 "Q": tags.CHAR_UP_Q, "R": tags.CHAR_UP_R,
                 "S": tags.CHAR_UP_S, "T": tags.CHAR_UP_T,
                 "U": tags.CHAR_UP_U, "V": tags.CHAR_UP_V,
                 "W": tags.CHAR_UP_W, "X": tags.CHAR_UP_X,
                 "Y": tags.CHAR_UP_Y, "Z": tags.CHAR_UP_Z}

# Mapping CHAR tags to lower and upper alphabet letters
TAG_TO_LETTER = {tags.CHAR_A: "a", tags.CHAR_B: "b",
                 tags.CHAR_C: "c", tags.CHAR_D: "d",
                 tags.CHAR_E: "e", tags.CHAR_F: "f",
                 tags.CHAR_G: "g", tags.CHAR_H: "h",
                 tags.CHAR_I: "i", tags.CHAR_J: "j",
                 tags.CHAR_K: "k", tags.CHAR_L: "l",
                 tags.CHAR_M: "m", tags.CHAR_N: "n",
                 tags.CHAR_O: "o", tags.CHAR_P: "p",
                 tags.CHAR_Q: "q", tags.CHAR_R: "r",
                 tags.CHAR_S: "s", tags.CHAR_T: "t",
                 tags.CHAR_U: "u", tags.CHAR_V: "v",
                 tags.CHAR_W: "w", tags.CHAR_X: "x",
                 tags.CHAR_Y: "y", tags.CHAR_Z: "z",
                 tags.CHAR_UP_A: "A", tags.CHAR_UP_B: "B",
                 tags.CHAR_UP_C: "C", tags.CHAR_UP_D: "D",
                 tags.CHAR_UP_E: "E", tags.CHAR_UP_F: "F",
                 tags.CHAR_UP_G: "G", tags.CHAR_UP_H: "H",
                 tags.CHAR_UP_I: "I", tags.CHAR_UP_J: "J",
                 tags.CHAR_UP_K: "K", tags.CHAR_UP_L: "L",
                 tags.CHAR_UP_M: "M", tags.CHAR_UP_N: "N",
                 tags.CHAR_UP_O: "O", tags.CHAR_UP_P: "P",
                 tags.CHAR_UP_Q: "Q", tags.CHAR_UP_R: "R",
                 tags.CHAR_UP_S: "S", tags.CHAR_UP_T: "T",
                 tags.CHAR_UP_U: "U", tags.CHAR_UP_V: "V",
                 tags.CHAR_UP_W: "W", tags.CHAR_UP_X: "X",
                 tags.CHAR_UP_Y: "Y", tags.CHAR_UP_Z: "Z"}



import unittest


class TestImports(unittest.TestCase):

    def test_import_funcs(self):
        try:
            # import functions
            from paradict import encode
            from paradict import decode
            from paradict import encode_into
            from paradict import decode_from
            from paradict import pack
            from paradict import unpack
            from paradict import pack_into
            from paradict import unpack_from
            from paradict import scan
            from paradict import forge_bin
            from paradict import stringify_bin
            from paradict import is_valid
            from paradict import split_kv
        except ImportError:
            self.assertTrue(False)

    def test_import_classes(self):
        try:
            # import classes
            from paradict import Encoder
            from paradict import Decoder
            from paradict import Packer
            from paradict import Unpacker
            from paradict import TypeRef
            from paradict import Validator
            from paradict import Datatype
        except ImportError:
            self.assertTrue(False)


if __name__ == "__main__":
    unittest.main()



"""binary Paradict: 256 tags from NOP (`\\x00`) to END (`\\xff`)"""


# Nop
NOP = b'\x00'

# ----------- CONTAINERS -------------

# Dict (container)
DICT = b'\x01'
DICT_EMPTY = b'\x02'

# List (container)
LIST = b'\x03'
LIST_EMPTY = b'\x04'

# Set (container)
SET = b'\x05'
SET_EMPTY = b'\x06'

# Extension Object (container)
OBJ = b'\x07'
OBJ_EMPTY = b'\x08'

# ----------- GRID -------------

# Grid (math container)
GRID = b'\x09'
GRID_DIV = b'\x0a'
GRID_EMPTY = b'\x0b'

# ----------- NULL AND BOOL -------------

# Null
NULL = b'\x0c'

# Bool
BOOL_TRUE = b'\x0d'
BOOL_FALSE = b'\x0e'

# ----------- COMPLEX NUMBER -------------

# <tag, real, imaginary>
COMPLEX = b'\x0f'

# ----------- RESERVED TAGS -------------

XA = b'\x10'
XB = b'\x11'
XC = b'\x12'
XD = b'\x13'
XE = b'\x14'

# ----------- DATE -------------

# <tag, year_delta, day_delta>
DATE = b'\x15'

# ----------- TIME -------------

# <tag, nanoseconds_leading_digits, trailing_zeros>
TIME = b'\x16'
# <tag, nanoseconds_leading_digits, trailing_zeros, utc_offset_minutes>
TIME_EXT = b'\x17'

# ----------- DATETIME -------------

# <tag, year_delta, nanoseconds_delta_leading_digits, trailing_zeros>
DATETIME = b'\x18'
# <tag, year_delta, nanoseconds_delta_leading_digits, trailing_zeros, utc_offset_minutes>
DATETIME_EXT = b'\x19'

# ----------- BIN OCT HEX INTEGER NOTATION -------------

# <tag, integer>
RADIX_BIN = b'\x1a'
# <tag, leading_zeros, integer>
RADIX_BIN_EXT = b'\x1b'

# <tag, integer>
RADIX_OCT = b'\x1c'
# <tag, leading_zeros, integer>
RADIX_OCT_EXT = b'\x1d'

# <tag, integer>
RADIX_HEX = b'\x1e'
# <tag, leading_zeros, integer>
RADIX_HEX_EXT = b'\x1f'

# ----------- FLOATING-POINT NUMBER-------------

# <tag, char> where char is a letter (n, x, y, z) for NaN, +inf, -inf, and -0
FLOAT_MISC = b'\x20'

# <tag, left_significand>
FLOAT_1 = b'\x21'
# <tag, left_significand, exponent>
FLOAT_1_EXT = b'\x22'

# <tag, left_significand, right_significand>
FLOAT_2 = b'\x23'
# <tag, left_significand, right_significand, exponent>
FLOAT_2_EXT = b'\x24'

# <tag, left_significand, leading_zeros, right_significand>
FLOAT_3 = b'\x25'
# <tag, left_significand, leading_zeros, right_significand, exponent>
FLOAT_3_EXT = b'\x26'

# ----------- BINARY -------------

# Bin
BIN_EMPTY = b'\x27'
BIN_SHORT = b'\x28'
BIN_MEDIUM = b'\x29'
BIN_LONG = b'\x2a'
BIN_BIG = b'\x2b'
BIN_HEAVY = b'\x2c'

# ----------- INTEGER -------------

# Positive Integer
PINT_8 = b'\x2d'
PINT_16 = b'\x2e'
PINT_24 = b'\x2f'
PINT_32 = b'\x30'
PINT_40 = b'\x31'
PINT_48 = b'\x32'
PINT_56 = b'\x33'
PINT_64 = b'\x34'
PINT_BIG = b'\x35'
PINT_HEAVY = b'\x36'

# Negative Integer
NINT_8 = b'\x37'
NINT_16 = b'\x38'
NINT_24 = b'\x39'
NINT_32 = b'\x3a'
NINT_40 = b'\x3b'
NINT_48 = b'\x3c'
NINT_56 = b'\x3d'
NINT_64 = b'\x3e'
NINT_BIG = b'\x3f'
NINT_HEAVY = b'\x40'

# ----------- STRING -------------

# String
STR_8 = b'\x41'
STR_16 = b'\x42'
STR_24 = b'\x43'
STR_32 = b'\x44'
STR_40 = b'\x45'
STR_48 = b'\x46'
STR_56 = b'\x47'
STR_64 = b'\x48'
STR_72 = b'\x49'
STR_80 = b'\x4a'
STR_88 = b'\x4b'
STR_96 = b'\x4c'
STR_104 = b'\x4d'
STR_112 = b'\x4e'
STR_120 = b'\x4f'
STR_128 = b'\x50'
STR_136 = b'\x51'
STR_144 = b'\x52'
STR_152 = b'\x53'
STR_160 = b'\x54'
STR_168 = b'\x55'
STR_176 = b'\x56'
STR_184 = b'\x57'
STR_192 = b'\x58'
STR_200 = b'\x59'
STR_208 = b'\x5a'
STR_216 = b'\x5b'
STR_224 = b'\x5c'
STR_232 = b'\x5d'
STR_240 = b'\x5e'
STR_248 = b'\x5f'
STR_256 = b'\x60'
STR_EMPTY = b'\x61'
STR_SHORT = b'\x62'
STR_MEDIUM = b'\x63'
STR_LONG = b'\x64'
STR_BIG = b'\x65'
STR_HEAVY = b'\x66'

# ----------- ALPHABET -------------

# a to z
CHAR_A = b'\x67'
CHAR_B = b'\x68'
CHAR_C = b'\x69'
CHAR_D = b'\x6a'
CHAR_E = b'\x6b'
CHAR_F = b'\x6c'
CHAR_G = b'\x6d'
CHAR_H = b'\x6e'
CHAR_I = b'\x6f'
CHAR_J = b'\x70'
CHAR_K = b'\x71'
CHAR_L = b'\x72'
CHAR_M = b'\x73'
CHAR_N = b'\x74'
CHAR_O = b'\x75'
CHAR_P = b'\x76'
CHAR_Q = b'\x77'
CHAR_R = b'\x78'
CHAR_S = b'\x79'
CHAR_T = b'\x7a'
CHAR_U = b'\x7b'
CHAR_V = b'\x7c'
CHAR_W = b'\x7d'
CHAR_X = b'\x7e'
CHAR_Y = b'\x7f'
CHAR_Z = b'\x80'

# A to Z (uppercase)
CHAR_UP_A = b'\x81'
CHAR_UP_B = b'\x82'
CHAR_UP_C = b'\x83'
CHAR_UP_D = b'\x84'
CHAR_UP_E = b'\x85'
CHAR_UP_F = b'\x86'
CHAR_UP_G = b'\x87'
CHAR_UP_H = b'\x88'
CHAR_UP_I = b'\x89'
CHAR_UP_J = b'\x8a'
CHAR_UP_K = b'\x8b'
CHAR_UP_L = b'\x8c'
CHAR_UP_M = b'\x8d'
CHAR_UP_N = b'\x8e'
CHAR_UP_O = b'\x8f'
CHAR_UP_P = b'\x90'
CHAR_UP_Q = b'\x91'
CHAR_UP_R = b'\x92'
CHAR_UP_S = b'\x93'
CHAR_UP_T = b'\x94'
CHAR_UP_U = b'\x95'
CHAR_UP_V = b'\x96'
CHAR_UP_W = b'\x97'
CHAR_UP_X = b'\x98'
CHAR_UP_Y = b'\x99'
CHAR_UP_Z = b'\x9a'

# ----------- CONSTANTS INTEGERS -------------

# Const 0 to 99
CONST_0 = b'\x9b'
CONST_1 = b'\x9c'
CONST_2 = b'\x9d'
CONST_3 = b'\x9e'
CONST_4 = b'\x9f'
CONST_5 = b'\xa0'
CONST_6 = b'\xa1'
CONST_7 = b'\xa2'
CONST_8 = b'\xa3'
CONST_9 = b'\xa4'
CONST_10 = b'\xa5'
CONST_11 = b'\xa6'
CONST_12 = b'\xa7'
CONST_13 = b'\xa8'
CONST_14 = b'\xa9'
CONST_15 = b'\xaa'
CONST_16 = b'\xab'
CONST_17 = b'\xac'
CONST_18 = b'\xad'
CONST_19 = b'\xae'
CONST_20 = b'\xaf'
CONST_21 = b'\xb0'
CONST_22 = b'\xb1'
CONST_23 = b'\xb2'
CONST_24 = b'\xb3'
CONST_25 = b'\xb4'
CONST_26 = b'\xb5'
CONST_27 = b'\xb6'
CONST_28 = b'\xb7'
CONST_29 = b'\xb8'
CONST_30 = b'\xb9'
CONST_31 = b'\xba'
CONST_32 = b'\xbb'
CONST_33 = b'\xbc'
CONST_34 = b'\xbd'
CONST_35 = b'\xbe'
CONST_36 = b'\xbf'
CONST_37 = b'\xc0'
CONST_38 = b'\xc1'
CONST_39 = b'\xc2'
CONST_40 = b'\xc3'
CONST_41 = b'\xc4'
CONST_42 = b'\xc5'
CONST_43 = b'\xc6'
CONST_44 = b'\xc7'
CONST_45 = b'\xc8'
CONST_46 = b'\xc9'
CONST_47 = b'\xca'
CONST_48 = b'\xcb'
CONST_49 = b'\xcc'
CONST_50 = b'\xcd'
CONST_51 = b'\xce'
CONST_52 = b'\xcf'
CONST_53 = b'\xd0'
CONST_54 = b'\xd1'
CONST_55 = b'\xd2'
CONST_56 = b'\xd3'
CONST_57 = b'\xd4'
CONST_58 = b'\xd5'
CONST_59 = b'\xd6'
CONST_60 = b'\xd7'
CONST_61 = b'\xd8'
CONST_62 = b'\xd9'
CONST_63 = b'\xda'
CONST_64 = b'\xdb'
CONST_65 = b'\xdc'
CONST_66 = b'\xdd'
CONST_67 = b'\xde'
CONST_68 = b'\xdf'
CONST_69 = b'\xe0'
CONST_70 = b'\xe1'
CONST_71 = b'\xe2'
CONST_72 = b'\xe3'
CONST_73 = b'\xe4'
CONST_74 = b'\xe5'
CONST_75 = b'\xe6'
CONST_76 = b'\xe7'
CONST_77 = b'\xe8'
CONST_78 = b'\xe9'
CONST_79 = b'\xea'
CONST_80 = b'\xeb'
CONST_81 = b'\xec'
CONST_82 = b'\xed'
CONST_83 = b'\xee'
CONST_84 = b'\xef'
CONST_85 = b'\xf0'
CONST_86 = b'\xf1'
CONST_87 = b'\xf2'
CONST_88 = b'\xf3'
CONST_89 = b'\xf4'
CONST_90 = b'\xf5'
CONST_91 = b'\xf6'
CONST_92 = b'\xf7'
CONST_93 = b'\xf8'
CONST_94 = b'\xf9'
CONST_95 = b'\xfa'
CONST_96 = b'\xfb'
CONST_97 = b'\xfc'
CONST_98 = b'\xfd'
CONST_99 = b'\xfe'

# ----------- -------------

# The End !
END = b'\xff'



"""High-level functions to serialize Python dict in Paradict binary/text format"""
from paradict.serializer.encoder import Encoder
from paradict.serializer.packer import Packer
from paradict import const


__all__ = ["encode", "pack"]


def encode(data, *, mode=const.DATA_MODE, type_ref=None,
           bin_to_text=True, root_dir=None, attachments_dir="attachments"):
    """
    Serialize a Python dict object with the Paradict binary format

    [param]
    - data: Python dict
    - mode: either const.DATA_MODE or const.CONFIG_MODE. Defaults to DATA_MODE.
    - type_ref: optional TypeRef object
    - bin_to_text: boolean to tell whether bin data should be converted into text or not
    - root_dir: root directory in which the attachments dir is supposed to be
    - attachments_dir: attachments directory. This is a path that is relative to the root dir.
     Note that relative paths should use a slash as separator.

    [return]
    Return a string in the Paradict text format
    """
    encoder = Encoder(mode=mode, type_ref=type_ref,
                      bin_to_text=bin_to_text, root_dir=root_dir,
                      attachments_dir=attachments_dir)
    lines = list()
    for r in encoder.encode(data):
        lines.append(r)
    return "\n".join(lines)


def pack(data, *, type_ref=None, dict_only=False):
    """
    Serialize a Python dict object with the Paradict binary format

    [param]
    - data: Python data object
    - type_ref: optional TypeRef object
    - dict_only: boolean to enforce dict as root

    [return]
    Return a Python bytes object packed in the Paradict binary format
    """
    packer = Packer(type_ref=type_ref, dict_only=dict_only)
    buffer = bytearray()
    for r in packer.pack(data):
        buffer.extend(r)
    return bytes(buffer)



import math
from paradict.typeref import TypeRef
from paradict import tags
from paradict.const import Datatype
from paradict.tags.mappings import SIZE_TO_PINT, \
    SIZE_TO_NINT, LETTER_TO_TAG, SIZE_TO_STR
from paradict import errors, misc, xtypes

__all__ = ["Packer"]


class Packer:
    """
    Class to convert some binary Python dict into Paradict binary format
    """
    def __init__(self, type_ref=None, dict_only=False, auto_index=False):
        """
        Init

        [param]
        - type_ref: optional TypeRef object
        - dict_only: boolean to enforce dict as root
        """
        self._type_ref = type_ref if type_ref else TypeRef()
        self._dict_only = dict_only
        self._auto_index = auto_index
        self._index_dict = dict()
        self._dict_counter = 0
        self._is_dict = False
        self._field_size = 0
        self._converters = {Datatype.DICT: self._pack_dict,
                            Datatype.LIST: self._pack_list,
                            Datatype.SET: self._pack_set,
                            Datatype.OBJ: self._pack_obj,
                            Datatype.GRID: self._pack_grid,
                            Datatype.BOOL: self._pack_bool,
                            Datatype.STR: self._pack_str,
                            Datatype.BIN: self._pack_bin,
                            Datatype.INT: self._pack_int,
                            Datatype.FLOAT: self._pack_float,
                            Datatype.COMPLEX: self._pack_complex,
                            Datatype.DATETIME: self._pack_datetime,
                            Datatype.DATE: self._pack_date,
                            Datatype.TIME: self._pack_time}

    @property
    def type_ref(self):
        return self._type_ref

    @type_ref.setter
    def type_ref(self, val):
        self._type_ref = val

    @property
    def dict_only(self):
        return self._dict_only

    @property
    def auto_index(self):
        return self._auto_index

    @property
    def index_dict(self):
        """Index dictionary for when the root container is a dict"""
        return self._index_dict

    def pack(self, data):
        """Generator for iteratively packing data by yielding bytes datum forged in
         Paradict binary format"""
        # clear variables
        self._dict_counter = 0
        self._field_size = 0
        self._index_dict = dict()
        # pack data
        data = self._type_ref.adapt(data)
        self._is_dict = type(data) in self._type_ref.dict_types
        if self._dict_only and not self._is_dict:
            msg = "The root data structure should be a dict"
            raise errors.Error(msg)
        yield from self._pack(data)

    def _pack(self, data):
        if data is None:
            yield from self._pack_null(data)
            return
        datatype = type(data)
        type_name = self._type_ref.check(datatype)
        try:
            r = self._converters[type_name](data)
        except KeyError as e:
            msg = "Unknown datatype '{}'".format(datatype)
            raise errors.Error(msg)
        yield from r

    def _pack_dict(self, data):
        if not data:
            yield tags.DICT_EMPTY
            return
        self._dict_counter += 1
        # encode dict head
        yield tags.DICT
        # iterate dict
        for key, val in data.items():
            key, val = self._type_ref.adapt(key), self._type_ref.adapt(val)
            if self._auto_index and self._is_dict and self._dict_counter == 1:
                key_size = val_size = 0
                for r in self._pack(key):
                    key_size += len(r)
                    yield r
                for r in self._pack(val):
                    val_size += len(r)
                    yield r
                start = key_size + 1 if self._field_size == 0 else self._field_size + 1
                self._field_size += key_size + val_size
                stop = self._field_size + 1
                self._index_dict[key] = slice(start, stop)
            else:
                yield from self._pack(key)
                yield from self._pack(val)
        self._dict_counter -= 1
        # encode dict tail
        yield tags.END

    def _pack_list(self, data):
        if not data:
            yield tags.LIST_EMPTY
            return
        # encode list head
        yield tags.LIST
        # iterate list
        for item in data:
            item = self._type_ref.adapt(item)
            yield from self._pack(item)
        # encode list tail
        yield tags.END

    def _pack_set(self, data):
        if not data:
            yield tags.SET_EMPTY
            return
        # encode set head
        yield tags.SET
        # iterate set
        for item in data:
            item = self._type_ref.adapt(item)
            if self._type_ref.check(type(item)) in (Datatype.DICT,
                                                    Datatype.LIST,
                                                    Datatype.SET,
                                                    Datatype.OBJ,
                                                    Datatype.GRID):
                msg = "The set container can't contain a dict, list, set, obj, or grid"
                raise errors.Error(msg)
            yield from self._pack(item)
        # encode set tail
        yield tags.END

    def _pack_obj(self, data):
        if not data:
            yield tags.OBJ_EMPTY
            return
        # encode obj head
        yield tags.OBJ
        # iterate obj
        for key, val in data.items():
            key, val = self._type_ref.adapt(key), self._type_ref.adapt(val)
            yield from self._pack(key)
            yield from self._pack(val)
        # encode obj tail
        yield tags.END

    def _pack_grid(self, data):
        if not data:
            yield tags.GRID_EMPTY
            return
        # encode grid head
        yield tags.GRID
        is_first_row = True
        row_size = 0
        # iterate grid
        for row in data:
            row = self._type_ref.adapt(row)
            if not row_size:
                row_size = len(row)
            if len(row) != row_size:
                raise errors.Error("Inconsistent grid")
            if type(row) not in self._type_ref.list_types:
                msg = "Unknown row type {}".format(type(row))
                raise errors.Error(msg)
            # iterate over the contents of the row
            for item in row:
                item = self._type_ref.adapt(item)
                typename = self._type_ref.check(type(item))
                if typename not in (Datatype.INT, Datatype.FLOAT, Datatype.COMPLEX):
                    msg = "A grid should be built with int, float, or complex numbers"
                    raise errors.Error(msg)
                # process int, float
                if typename == Datatype.INT:
                    yield from self._pack_int(item)
                elif typename == Datatype.FLOAT:
                    yield from self._pack_float(item)
                elif typename == Datatype.COMPLEX:
                    yield from self._pack_complex(item)
            # add grid div if this is the first row
            if is_first_row and len(data) > 1:
                yield tags.GRID_DIV
                is_first_row = False
        # encode grid tail
        yield tags.END

    def _pack_null(self, data):
        yield tags.NULL

    def _pack_bool(self, data):
        yield pack_bool(data)

    def _pack_str(self, data):
        yield pack_str(data)

    def _pack_bin(self, data):
        yield pack_bin(data)

    def _pack_int(self, data):
        if isinstance(data, xtypes.HexInt):
            yield from self._pack_hex_int(data)
            return
        if isinstance(data, xtypes.BinInt):
            yield from self._pack_bin_int(data)
            return
        if isinstance(data, xtypes.OctInt):
            yield from self._pack_oct_int(data)
            return
        yield pack_int(data)

    def _pack_hex_int(self, data):
        yield pack_hex_int(data)

    def _pack_oct_int(self, data):
        yield pack_oct_int(data)

    def _pack_bin_int(self, data):
        yield pack_bin_int(data)

    def _pack_float(self, data):
        yield pack_float(data)

    def _pack_complex(self, data):
        yield pack_complex(data)

    def _pack_datetime(self, data):
        yield pack_datetime(data)

    def _pack_date(self, data):
        yield pack_date(data)

    def _pack_time(self, data):
        yield pack_time(data)


def pack_bool(val):
    if val:
        return tags.BOOL_TRUE
    else:
        return tags.BOOL_FALSE


def pack_str(val):
    if not val:
        return tags.STR_EMPTY
    # const letters
    if len(val) == 1:
        try:
            return LETTER_TO_TAG[val]
        except KeyError as e:
            pass
    data = val.encode("utf-8")
    # 8 to 64 bits
    size = len(data)
    try:
        tag_byte = SIZE_TO_STR[size]
    except KeyError as e:
        pass
    else:
        return misc.forge_bin(tag_byte, data)
    # short
    if size <= 2**8:
        return misc.forge_bin(tags.STR_SHORT, size-1, data)
    # medium
    if size <= 2**(8*2):
        return misc.forge_bin(tags.STR_MEDIUM, size-1, data)
    # long
    if size <= 2**(8*3):
        return misc.forge_bin(tags.STR_LONG, size-1, data)
    # big
    if size <= 2**(8*4):
        return misc.forge_bin(tags.STR_BIG, size-1, data)
    # heavy
    if size <= 2**(8*5):
        return misc.forge_bin(tags.STR_HEAVY, size-1, data)
    # oveeeer heavy
    msg = "The string to pack is over heavy !"
    raise errors.Error(msg)


def pack_bin(val):
    if not val:
        return tags.BIN_EMPTY
    # 8 to 64 bits
    size = len(val)
    # short
    if size <= 2**8:
        return misc.forge_bin(tags.BIN_SHORT, size-1, val)
    # medium
    if size <= 2**(8*2):
        return misc.forge_bin(tags.BIN_MEDIUM, size-1, val)
    # long
    if size <= 2**(8*3):
        return misc.forge_bin(tags.BIN_LONG, size-1, val)
    # big
    if size <= 2**(8*4):
        return misc.forge_bin(tags.BIN_BIG, size-1, val)
    # heavy
    if size <= 2**(8*5):
        return misc.forge_bin(tags.BIN_HEAVY, size-1, val)
    # oveeeer heavy
    msg = "The bin to pack is over heavy !"
    raise errors.Error(msg)


def pack_int(val):
    # yield integer
    if val >= 0:
        return pack_pint(val)
    else:
        return pack_nint(val)


def pack_hex_int(val):
    leading_zeros = misc.count_leading_zeros(val)
    if leading_zeros == 0:
        return misc.forge_bin(tags.RADIX_HEX, pack_int(val))
    return misc.forge_bin(tags.RADIX_HEX_EXT,
                          pack_int(leading_zeros),
                          pack_int(val))


def pack_oct_int(val):
    leading_zeros = misc.count_leading_zeros(val)
    if leading_zeros == 0:
        return misc.forge_bin(tags.RADIX_OCT, pack_int(val))
    return misc.forge_bin(tags.RADIX_OCT_EXT,
                          pack_int(leading_zeros),
                          pack_int(val))


def pack_bin_int(val):
    leading_zeros = misc.count_leading_zeros(val)
    if leading_zeros == 0:
        return misc.forge_bin(tags.RADIX_BIN, pack_int(val))
    return misc.forge_bin(tags.RADIX_BIN_EXT,
                          pack_int(leading_zeros),
                          pack_int(val))


def pack_pint(val):
    size = misc.calc_uint_bytes(val)  # size is the n of bytes
    if val < 100:
        tag = bytes([tags.CONST_0[0] + val])
        return misc.forge_bin(tag)
    if size <= 8:
        tag = SIZE_TO_PINT.get(size)
        return misc.forge_bin(tag, val)
    if size <= 256:
        tag = tags.PINT_BIG
        return misc.forge_bin(tag, size-1, val)
    if size <= 65536:
        tag = tags.PINT_HEAVY
        return misc.forge_bin(tag, size-1, val)
    msg = "Too large positive integer"
    raise errors.Error(msg)


def pack_nint(val):
    val = abs(val)
    size = misc.calc_uint_bytes(val)  # size is the n of bytes
    if size <= 8:
        tag = SIZE_TO_NINT.get(size)
        return misc.forge_bin(tag, val)
    if size <= 256:
        tag = tags.NINT_BIG
        return misc.forge_bin(tag, size-1, val)
    if size <= 65536:
        tag = tags.NINT_HEAVY
        return misc.forge_bin(tag, size-1, val)
    msg = "Too large negative integer"
    raise errors.Error(msg)


def pack_float(val):
    # process special float values
    if val == 0:
        if math.copysign(1, val) == -1:
            return misc.forge_bin(tags.FLOAT_MISC, tags.CHAR_Z)
        else:
            return _pack_regular_float(val)
    elif val == float("+inf"):
        return misc.forge_bin(tags.FLOAT_MISC, tags.CHAR_X)
    elif val == float("-inf"):
        return misc.forge_bin(tags.FLOAT_MISC, tags.CHAR_Y)
    elif val == float("NaN"):
        return misc.forge_bin(tags.FLOAT_MISC, tags.CHAR_N)
    else:
        return _pack_regular_float(val)


def _pack_regular_float(val):
    info = misc.split_float(val)
    exponent = int(info.exponent)
    left_significand = int(info.sign + info.left_significand)
    right_significand = int(info.right_significand)
    leading_zeros = misc.count_leading_zeros(info.right_significand)
    if right_significand == 0:
        r = _pack_float_1(left_significand, exponent)
    elif right_significand != 0 and leading_zeros == 0:
        r = _pack_float_2(left_significand, right_significand, exponent)
    elif right_significand != 0 and leading_zeros != 0:
        r = _pack_float_3(left_significand, leading_zeros,
                          right_significand, exponent)
    else:
        msg = "Inconsistent float value"
        raise errors.Error(msg)
    return r


def _pack_float_1(left_significand, exponent):
    if exponent == 0:
        r = misc.forge_bin(tags.FLOAT_1,
                           pack_int(left_significand))
    else:
        r = misc.forge_bin(tags.FLOAT_1_EXT,
                           pack_int(left_significand),
                           pack_int(exponent))
    return r


def _pack_float_2(left_significand, right_significand, exponent):
    if exponent == 0:
        r = misc.forge_bin(tags.FLOAT_2,
                           pack_int(left_significand),
                           pack_int(right_significand))
    else:
        r = misc.forge_bin(tags.FLOAT_2_EXT,
                           pack_int(left_significand),
                           pack_int(right_significand),
                           pack_int(exponent))
    return r


def _pack_float_3(left_significand, leading_zeros, right_significand, exponent):
    if exponent == 0:
        r = misc.forge_bin(tags.FLOAT_3,
                           pack_int(left_significand),
                           pack_int(leading_zeros),
                           pack_int(right_significand))
    else:
        r = misc.forge_bin(tags.FLOAT_3_EXT,
                           pack_int(left_significand),
                           pack_int(leading_zeros),
                           pack_int(right_significand),
                           pack_int(exponent))
    return r


def pack_complex(val):
    real, imag = val.real, val.imag
    real = int(real) if misc.is_whole_number(real) else real
    imag = int(imag) if misc.is_whole_number(imag) else imag
    packed_real = pack_int(real) if isinstance(real, int) else pack_float(real)
    packed_imag = pack_int(imag) if isinstance(imag, int) else pack_float(imag)
    return misc.forge_bin(tags.COMPLEX,
                          packed_real, packed_imag)


def pack_datetime(val):
    year_delta, ns_delta, trailing_zeros, tz = misc.deconstruct_datetime(val)
    if tz is None:
        r = misc.forge_bin(tags.DATETIME,
                           pack_int(year_delta),
                           pack_int(ns_delta),
                           pack_int(trailing_zeros))
    else:
        r = misc.forge_bin(tags.DATETIME_EXT,
                           pack_int(year_delta),
                           pack_int(ns_delta),
                           pack_int(trailing_zeros),
                           pack_int(tz))
    return r


def pack_date(val):
    year_delta, days_delta = misc.deconstruct_date(val)
    r = misc.forge_bin(tags.DATE,
                       pack_int(year_delta),
                       pack_int(days_delta))
    return r


def pack_time(val):
    ns_delta, trailing_zeros, tz = misc.deconstruct_time(val)
    if tz is None:
        r = misc.forge_bin(tags.TIME,
                           pack_int(ns_delta),
                           pack_int(trailing_zeros))
    else:
        r = misc.forge_bin(tags.TIME_EXT,
                           pack_int(ns_delta),
                           pack_int(trailing_zeros),
                           pack_int(tz))
    return r



import os
import re
import base64
from textwrap import dedent
from collections import namedtuple
from paradict import errors, misc, const, xtypes
from paradict.const import Datatype
from paradict.typeref import TypeRef

Context = namedtuple("Context", ["name", "collection", "indents"])


class Encoder:
    """Convert a Python dictionary object to Paradict text format"""
    def __init__(self, mode=const.DATA_MODE, type_ref=None,
                 bin_to_text=True, root_dir=None, attachments_dir="attachments"):
        """
        Init

        [param]
        - mode: either const.DATA_MODE or const.CONFIG_MODE. Defaults to DATA_MODE.
        - type_ref: optional TypeRef object
        - bin_to_text: boolean to tell whether bin data should be embedded as base16 or
        stored in a linked file
        - root_dir: root directory in which the attachments dir is supposed to be
        - attachments_dir: attachments directory. This is a path that is relative to the root dir.
            Note that relative paths should use a slash as separator.
        """
        self._mode = mode
        self._type_ref = type_ref if type_ref else TypeRef()
        self._bin_to_text = bin_to_text
        self._root_dir = root_dir
        self._attachments_dir = attachments_dir if attachments_dir else ""
        self._stack = list()
        self._converters = {Datatype.DICT: self._encode_dict,
                            Datatype.LIST: self._encode_list,
                            Datatype.SET: self._encode_set,
                            Datatype.OBJ: self._encode_obj,
                            Datatype.GRID: self._encode_grid,
                            Datatype.BOOL: self._encode_bool,
                            Datatype.STR: self._encode_str,
                            Datatype.BIN: self._encode_bin,
                            Datatype.INT: self._encode_int,
                            Datatype.FLOAT: self._encode_float,
                            Datatype.COMPLEX: self._encode_complex,
                            Datatype.DATETIME: self._encode_datetime,
                            Datatype.DATE: self._encode_date,
                            Datatype.TIME: self._encode_time}

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, val):
        self._mode = val

    @property
    def type_ref(self):
        return self._type_ref

    @type_ref.setter
    def type_ref(self, val):
        self._type_ref = val

    @property
    def bin_to_text(self):
        return self._bin_to_text

    @bin_to_text.setter
    def bin_to_text(self, val):
        self._bin_to_text = val

    @property
    def root_dir(self):
        return self._root_dir

    @root_dir.setter
    def root_dir(self, val):
        self._root_dir = val

    @property
    def attachments_dir(self):
        return self._attachments_dir

    @attachments_dir.setter
    def attachments_dir(self, val):
        self._attachments_dir = val

    def encode(self, data):
        """Generator for iteratively encoding data by yielding lines of Paradict text format"""
        data = self._type_ref.adapt(data)
        if type(data) not in self._type_ref.dict_types:
            msg = "The root data structure should be a dict"
            raise errors.Error(msg)
        g = self._encode(data)
        next(g)
        yield from g

    def _encode(self, data, indents=-1):
        if data is None:
            yield from self._encode_null(data, indents)
            return
        datatype = type(data)
        type_name = self._type_ref.check(datatype)
        try:
            converter = self._converters[type_name]
        except KeyError as e:
            msg = "Unknown datatype '{}'".format(datatype)
            raise errors.Error(msg)
        yield from converter(data, indents)

    def _check_key(self, key):
        type_name = self._type_ref.check(type(key))
        if (type_name == Datatype.STR
                or (self._mode == const.DATA_MODE
                    and type_name in (Datatype.INT, Datatype.FLOAT, Datatype.COMPLEX))):
            pass
        else:
            msg = ("In CONFIG_MODE, a key should be either str or raw"
                   " without whitespace. "
                   "In DATA_MODE, a key should be one of: str, "
                   "int, float, complex.")
            raise errors.Error(msg)
        if type_name == Datatype.STR:
            key = key.replace("\\", "\\\\")
            key = key.replace("\n", "\\n")
            r = '"{}"'.format(key)
        else:
            r = next(self._encode(key))
        # ensure that key doesn't contain whitespace or "=" symbol
        # if CONFIG_MODE is on
        if self._mode == const.CONFIG_MODE:
            r = r.strip("'").strip('"')
            if " " in r or "=" in r:
                msg = ("Keys in CONFIG_MODE shouldn't contain "
                       "a space or the equal sign (=).")
                raise errors.Error(msg)
        # replace newline with
        else:
            r = r.replace("\n", "\\n")
        return r

    def _encode_dict(self, data, indents):
        yield from self._encode_dict_and_obj("dict", data, indents)

    def _encode_obj(self, data, indents=0):
        yield from self._encode_dict_and_obj("obj", data, indents)

    def _encode_list(self, data, indents=0):
        indent_str = misc.make_indent_str(indents)
        yield indent_str + "(list)"
        # iterate list
        for val in data:
            val = self._type_ref.adapt(val)
            g = self._encode(val, indents+1)
            val = next(g)
            yield val
            yield from g

    def _encode_set(self, data, indents=0):
        indent_str = misc.make_indent_str(indents)
        yield indent_str + "(set)"
        # iterate set
        for val in data:
            val = self._type_ref.adapt(val)
            valid_types = (Datatype.INT, Datatype.FLOAT, Datatype.COMPLEX,
                           Datatype.STR, Datatype.BIN, Datatype.DATETIME,
                           Datatype.DATE, Datatype.TIME)
            if self._type_ref.check(type(val)) not in valid_types:

                msg = "The set container can contain these types: {}"
                msg = msg.format(", ".join([x.name for x in valid_types]))
                raise errors.Error(msg)
            g = self._encode(val, indents + 1)
            val = next(g)
            yield val
            yield from g

    def _encode_dict_and_obj(self, tag, data, indents=0):
        indent_str = misc.make_indent_str(indents)
        yield indent_str + "({})".format(tag)
        # iterate dict
        for key, val in data.items():
            key, val = self._type_ref.adapt(key), self._type_ref.adapt(val)
            # process key and val
            key = self._check_key(key)
            sep = ": " if self._mode == const.DATA_MODE else " = "
            g = self._encode(val, indents + 1)
            val = dedent(next(g))
            indent_str = misc.make_indent_str(indents + 1)
            yield "{indent}{key}{sep}{val}".format(indent=indent_str,
                                                   key=key, sep=sep,
                                                   val=val)
            yield from g

    def _encode_grid(self, data, indents=0):
        indent_str = misc.make_indent_str(indents)
        yield indent_str + "(grid)"
        # iterate grid
        n = 0
        rows = list()
        # stringify
        for row in data:
            # check row consistency
            if n and len(row) != n:
                raise errors.Error("Inconsistent grid size")
            else:
                n = len(row)
            # iterate over the row
            cache = list()
            for val in row:
                val = self._type_ref.adapt(val)
                type_name = self._type_ref.check(type(val))
                if type_name not in (Datatype.INT,
                                     Datatype.FLOAT, Datatype.COMPLEX):
                    msg = "A Grid should be made of numbers only (integer, float, complex)"
                    raise errors.Error(msg)
                g = self._encode(val, indents + 1)
                val = dedent(next(g))
                cache.append(val)
            rows.append(cache)
        pretty_grid = misc.prettify_grid(rows)
        # padding
        for row in pretty_grid:
            line = " ".join(row)
            indent_str = misc.make_indent_str(indents+1)
            yield indent_str + line

    def _encode_null(self, data, indents=0):
        indent_str = misc.make_indent_str(indents)
        yield indent_str + encode_null(data)

    def _encode_bool(self, data, indents=0):
        indent_str = misc.make_indent_str(indents)
        yield indent_str + encode_bool(data)

    def _encode_str(self, data, indents=0):
        #data = data.replace("\\", "\\\\")
        indent_str = misc.make_indent_str(indents)
        if "\n" in data:
            tag = "(raw)" if "\\" in data else "(text)"
            yield indent_str + tag
            indent_str = misc.make_indent_str(indents+1)
            for line in encode_multiline_str(data):
                yield indent_str + line
            yield indent_str + "---"
        else:
            val = encode_str(data)
            yield indent_str + val

    def _encode_bin(self, data, indents=0):
        indent_str = misc.make_indent_str(indents)
        if self._bin_to_text:
            yield indent_str + "(bin)"
            indent_str = misc.make_indent_str(indents + 1)
            for line in encode_bin(data):
                yield indent_str + line
        else:
            parts = misc.split_relative_path(self._attachments_dir)
            root_dir = self._root_dir if self._root_dir else os.getcwd()
            basename = misc.store_attachment(data, os.path.join(root_dir,
                                                                *parts))
            if self._attachments_dir:
                line = "load('{}/{}')".format(self._attachments_dir, basename)
            else:
                line = "load('{}')".format(basename)
            yield indent_str + line

    def _encode_int(self, data, indents=0):
        if isinstance(data, xtypes.HexInt):
            yield from self._encode_hex_int(data, indents)
            return
        if isinstance(data, xtypes.BinInt):
            yield from self._encode_bin_int(data, indents)
            return
        if isinstance(data, xtypes.OctInt):
            yield from self._encode_oct_int(data, indents)
            return
        indent_str = misc.make_indent_str(indents)
        r = encode_int(data)
        if len(r) > 42:
            yield indent_str + "(int)"
            indent_str = misc.make_indent_str(indents + 1)
            for line in misc.make_multiline(r):
                yield indent_str + line
        else:
            yield indent_str + r

    def _encode_hex_int(self, data, indents=0):
        indent_str = misc.make_indent_str(indents)
        r = encode_hex_int(data)
        if len(r) > 42:
            yield indent_str + "(int)"
            indent_str = misc.make_indent_str(indents + 1)
            for line in misc.make_multiline(r):
                yield indent_str + line
        else:
            yield indent_str + r

    def _encode_oct_int(self, data, indents=0):
        indent_str = misc.make_indent_str(indents)
        r = encode_oct_int(data)
        if len(r) > 42:
            yield indent_str + "(int)"
            indent_str = misc.make_indent_str(indents + 1)
            for line in misc.make_multiline(r):
                yield indent_str + line
        else:
            yield indent_str + r

    def _encode_bin_int(self, data, indents=0):
        indent_str = misc.make_indent_str(indents)
        r = encode_bin_int(data)
        if len(r) > 42:
            yield indent_str + "(int)"
            indent_str = misc.make_indent_str(indents + 1)
            for line in misc.make_multiline(r):
                yield indent_str + line
        else:
            yield indent_str + r

    def _encode_float(self, data, indents=0):
        indent_str = misc.make_indent_str(indents)
        r = encode_float(data)

        if len(r) > 42:
            yield indent_str + "(float)"
            indent_str = misc.make_indent_str(indents + 1)
            for line in misc.make_multiline(r):
                yield indent_str + line
        else:
            yield indent_str + r

    def _encode_complex(self, data, indents=0):
        indent_str = misc.make_indent_str(indents)
        yield indent_str + encode_complex(data)

    def _encode_datetime(self, data, indents=0):
        indent_str = misc.make_indent_str(indents)
        yield indent_str + encode_datetime(data)

    def _encode_date(self, data, indents=0):
        indent_str = misc.make_indent_str(indents)
        yield indent_str + encode_date(data)

    def _encode_time(self, data, indents=0):
        indent_str = misc.make_indent_str(indents)
        yield indent_str + encode_time(data)


def encode_null(val):
    return "null"


def encode_bool(val):
    if val:
        return "true"
    return "false"


def encode_bin(val):
    if not val:
        return list()
    r = base64.b16encode(val)
    s = r.decode("utf-8")
    return misc.prettify_base16(s)


def encode_str(val):
    quote = "'" if "\\" in val else '"'
    # indent_str = misc.make_indent_str(0)
    return "{quote}{data}{quote}".format(data=val, quote=quote)


def encode_multiline_str(val):
    return val.split("\n")


def encode_int(val):
    return misc.tidy_up_int(val, width=3)


def encode_hex_int(val):
    return misc.tidy_up_int(val, width=4)


def encode_oct_int(val):
    return misc.tidy_up_int(val, width=3)


def encode_bin_int(val):
    return misc.tidy_up_int(val, width=4)


def encode_float(val):
    return misc.tidy_up_float(val)


def encode_complex(val):
    real, imag = val.real, val.imag
    #real = int(real) if real.is_integer() else real
    #imag = int(imag) if imag.is_integer() else imag
    # tidy up real part
    if misc.is_whole_number(real):
        tidy_real = misc.tidy_up_int(int(real))
    else:
        tidy_real = misc.tidy_up_float(real)
    # tidy up imaginary part
    if misc.is_whole_number(imag):
        tidy_imag = misc.tidy_up_int(int(imag))
    else:
        tidy_imag = misc.tidy_up_float(imag)
    return "{}{}{}i".format(tidy_real, "" if imag < 0 else "+", tidy_imag)


def encode_datetime(val):
    value = val.isoformat()
    if value.endswith("+00:00"):
        value = value.replace("+00:00", "Z")
    return str(value)


def encode_date(val):
    """Parse date and time with the ISO 8601 format"""
    # date ISO format
    return str(val)


def encode_time(val):
    # time ISO format
    return str(val)
